﻿import cyan3d.Vector2D;
import cyan3d.Vector3D;

class cyan3d.Matrix3D
{
	var _e00:Number;
	var _e01:Number;
	var _e02:Number;
	var _e03:Number;
	var _e10:Number;
	var _e11:Number;
	var _e12:Number;
	var _e13:Number;
	var _e20:Number;
	var _e21:Number;
	var _e22:Number;
	var _e23:Number;
	var _e30:Number;
	var _e31:Number;
	var _e32:Number;
	var _e33:Number;

	static var toRadian:Number = (Math.PI/180.0);
	static var REPLACE = 0;
	static var PRE_CONCAT = 1;
	static var POST_CONCAT = 2;

    public function Matrix3D()
	{
		identity();
	}
 
	public function identity()
    {
       _e01 = _e02 = _e03 = 0.0;
       _e10 = _e12 = _e13 = 0.0;
       _e20 = _e21 = _e23 = 0.0;
       _e30 = _e31 = _e32 = 0.0;
       _e00 = 1.0;
       _e11 = 1.0;
       _e22 = 1.0;
	   _e33 = 1.0;
    }

	public function getPos():Vector3D
	{
		return(new Vector3D(_e03,_e13,_e23)); 
	}
	
	public function transform(v:Vector3D):Vector3D
    {
		var x:Number;
		var y:Number;
		var z:Number;
		x = v._x * _e00 + v._y * _e01 + v._z * _e02 + _e03;
		y = v._x * _e10 + v._y * _e11 + v._z * _e12 + _e13;
		z = v._x * _e20 + v._y * _e21 + v._z * _e22 + _e23;
		return(new Vector3D(x,y,z));
    }

	public function translate(x:Number, y:Number, z:Number)
	{
		_e03 += x;
		_e13 += y;
		_e23 += z;
	}

	public function scale(x:Number, y:Number, z:Number)
	{
		_e00 *= x;
		_e11 *= y;
		_e22 *= z;
	}
   
	public function scaleXYZ(s:Number)
	{
		_e00 *= s;
		_e11 *= s;
		_e22 *= s;
	}

	public function rotateX(theta:Number)
	{
		theta *= Matrix3D.toRadian;

		var theta_cos:Number = Math.cos(theta);
		var theta_sin:Number = Math.sin(theta);

		var m:Matrix3D = new Matrix3D();
		m._e11 = theta_cos;
		m._e12 = -theta_sin;
		m._e21 = theta_sin;
		m._e22 = theta_cos;
		this.multiply(m, Matrix3D.POST_CONCAT);
	}

	public function rotateY(theta:Number)
	{
		theta *= Matrix3D.toRadian;
		var theta_cos:Number = Math.cos(theta);
		var theta_sin:Number = Math.sin(theta);

		var m:Matrix3D = new Matrix3D();
		m._e00 = theta_cos;
		m._e02 = theta_sin;
		m._e20 = -theta_sin;
		m._e22 = theta_cos;
		this.multiply(m, Matrix3D.POST_CONCAT);
	}

	public function rotateZ(theta:Number)
	{
		theta *= Matrix3D.toRadian;
		var theta_cos:Number = Math.cos(theta);
		var theta_sin:Number = Math.sin(theta);
 
		var m:Matrix3D = new Matrix3D();
		m._e00 = theta_cos;
		m._e01 = -theta_sin;
		m._e10 = theta_sin;
		m._e11 = theta_cos;
		this.multiply(m, Matrix3D.POST_CONCAT);
	}

	public function multiply(matrix:Matrix3D,concat)
	{
		var m :Matrix3D = null;
		var m2:Matrix3D = null;
		if (concat == Matrix3D.PRE_CONCAT)
		{
			m = matrix;
			m2 = this;
		}
		else if (concat == Matrix3D.POST_CONCAT)
		{
			m = this;
			m2 = matrix;
		}
		else if (concat == Matrix3D.REPLACE)
		{
			m = this;
			m.identity();
			m2 = matrix;
		}

		var e00:Number;
		var e01:Number;
		var e02:Number;
		var e03:Number;
		
		var e10:Number;
		var e11:Number;
		var e12:Number;
		var e13:Number;

		var e20:Number;
		var e21:Number;
		var e22:Number;
		var e23:Number;

		e00 = m._e00 * m2._e00 + m._e10 * m2._e01 + m._e20 * m2._e02;
		e01 = m._e01 * m2._e00 + m._e11 * m2._e01 + m._e21 * m2._e02;
		e02 = m._e02 * m2._e00 + m._e12 * m2._e01 + m._e22 * m2._e02;
		e03 = m._e03 * m2._e00 + m._e13 * m2._e01 + m._e23 * m2._e02 + m2._e03;

		e10 = m._e00 * m2._e10 + m._e10 * m2._e11 + m._e20 * m2._e12;
		e11 = m._e01 * m2._e10 + m._e11 * m2._e11 + m._e21 * m2._e12;
		e12 = m._e02 * m2._e10 + m._e12 * m2._e11 + m._e22 * m2._e12;
		e13 = m._e03 * m2._e10 + m._e13 * m2._e11 + m._e23 * m2._e12 + m2._e13;

		e20 = m._e00 * m2._e20 + m._e10 * m2._e21 + m._e20 * m2._e22;
		e21 = m._e01 * m2._e20 + m._e11 * m2._e21 + m._e21 * m2._e22;
		e22 = m._e02 * m2._e20 + m._e12 * m2._e21 + m._e22 * m2._e22;
		e23 = m._e03 * m2._e20 + m._e13 * m2._e21 + m._e23 * m2._e22 + m2._e23;

		_e00 = e00;
		_e01 = e01;
		_e02 = e02;
		_e03 = e03;

		_e10 = e10;
		_e11 = e11;
		_e12 = e12;
		_e13 = e13;

		_e20 = e20;
		_e21 = e21;
		_e22 = e22;
		_e23 = e23;
	}
	
	public function perspective(fovy:Number, aspect:Number, zn:Number, zf:Number)
	{
		fovy = fovy * toRadian;
		var yScale:Number = 1 / Math.tan(fovy / 2);
		var xScale:Number = aspect * yScale;
		var m:Matrix3D = new Matrix3D();

		m._e00 = xScale;
		m._e01 = 0;
		m._e02 = 0;
		m._e03 = 0;
		
		m._e10 = 0;
		m._e11 = yScale;
		m._e12 = 0;
		m._e13 = 0;
		
		m._e20 = 0;
		m._e21 = 0;
		m._e22 = zf / (zf - zn);
		m._e23 = -zn * zf / (zf - zn);
		
		m._e30 = 0;
		m._e31 = 0;
		m._e32 = 1;
		m._e33 = 0;

		this.multiply(m, Matrix3D.POST_CONCAT);
	}

	//パースペクィブ2Dに
	public function transform2D(v:Vector3D, d:Number):Vector2D
	{
		var x:Number;
		var y:Number;
		var z:Number;

		x = v._x * _e00 + v._y * _e01 + v._z * _e02 + _e03;
		y = v._x * _e10 + v._y * _e11 + v._z * _e12 + _e13;
		z = v._x * _e20 + v._y * _e21 + v._z * _e22 + _e23;
  
		var x2:Number = d*x/z;
		var y2:Number = d*y/z;

		return(new Vector2D(x2,y2));
	}

    public function toString():String
    {
        return(new String(
			"("+_e00+","+_e01+","+_e02+","+_e03+")"+
			"("+_e10+","+_e11+","+_e12+","+_e13+")"+
			"("+_e20+","+_e21+","+_e22+","+_e23+")"+
			"("+_e30+","+_e31+","+_e32+","+_e33+")"));
    }
}
