﻿import cyan3d.*;

class cyan3d.Model3D
{
	var _world:Array = new Array();
	var _2d:Array = new Array();
	var _face:Array = new Array();
	var _material:Array = new Array();
	var _index:Array = new Array();
	var _line = 0;
	var _matrix:Matrix3D;
	var _trans:Vector3D;
	var _rotate:Vector3D;
	var _scale:Vector3D;

	public function Model3D() 
	{
		//変形行列
		_matrix = new Matrix3D();
		_trans = new Vector3D(0, 0, 0);
		_rotate = new Vector3D(0, 0, 0);
		_scale = new Vector3D(1, 1, 1);
	}
	
	public function init()
	{
		var v:Vector3D = new Vector3D();
		var v1:Vector3D = new Vector3D();
		var v2:Vector3D = new Vector3D();
		
		//法線計算
		for ( var i = 0; i < _face.length; i++ )
		{
			var face = _face[i];
			v  = _world[face._index[0]];
			v1 = _world[face._index[1]];
			v2 = _world[face._index[2]];
		
			face.setNormal(v,v1,v2);
			face.setShade(_material[face._material]);
		}
		//二次元座標
		_2d = new Array();
		for (var i = 0; i < _world.length; i++)
		{
			_2d.push(new Vector2D(0, 0));
		}
	}
	
	public function addM(r:Number,g:Number,b:Number)
	{
		_material.push(new Material(r, g, b));
	}

	public function addV(x, y, z)
	{
		_world.push(new Vector3D(x, y, z));
	}

	public function addF(index:Array,material)
	{
		_face.push(new Face3D(index,material));
	}
	
	public function draw(mc:MovieClip)
    {
		if ( _line > 0 )
		{
			mc.lineStyle(_line, 0x000000, 100); 
		}
		var v :Vector2D = new Vector2D();
		var v2:Vector2D = new Vector2D();
		for (var i = 0; i < _index.length; i++)
		{
			var face = _face[_index[i]];
			//隠面消去
			if ( !face._visible )
			{
				continue;
			}
			//三角形・四角形描画
			mc.beginFill(face._color, 100);
			v = _2d[face._index[0]];
			mc.moveTo(v._x, v._y);
			v2 = _2d[face._index[1]];
			mc.lineTo(v2._x, v2._y);
			v2 = _2d[face._index[2]];
			mc.lineTo(v2._x, v2._y);
			if ( face._index.length >= 4 )
			{
				v2 = _2d[face._index[3]];
				mc.lineTo(v2._x, v2._y);
			}
			mc.lineTo(v._x, v._y);
			mc.endFill();
		}
    }

	public function transform(camera:Camera3D)
    {
		//変形
		_matrix = new Matrix3D();
		_matrix.rotateX(_rotate._x);
		_matrix.rotateY(_rotate._y);
		_matrix.rotateZ(_rotate._z);
		_matrix.scale(_scale._x, _scale._y, _scale._z); 
		_matrix.translate(_trans._x, _trans._y, _trans._z); 
		
		var m:Matrix3D = new Matrix3D();
		m.multiply(camera.viewMatrix(), Matrix3D.REPLACE);
		m.multiply(_matrix, Matrix3D.PRE_CONCAT);
		//二次元へ
		for (var i = 0; i < _2d.length; i++) 
        {
			_2d[i] = m.transform2D(_world[i], 400);
		}
		hiddenSurfaceRemoval(camera._position);
		zSortAndNearClipping(m,camera);
	}
	
	private function hiddenSurfaceRemoval(pos:Vector3D)
	{
		//隠面消去
		for (var i = 0; i < _face.length; i++) 
        {
			var face = _face[i];
			var n:Vector3D = _matrix.transform(face._normal);
			var v:Vector3D = _matrix.transform(_world[face._index[0]]);
			v._x = v._x - pos._x;
			v._y = v._y - pos._y;
			v._z = v._z - pos._z;

			if ( n.dotProduct(v) > 0.0)
			{
				face._visible = false;
			}
			else
			{
				face._visible = true;
			}
		}
	}
	
	private function zSortAndNearClipping(m:Matrix3D,camera:Camera3D)
	{
		var i;
		var faces:Array = new Array();
		
		//Z値を追加
		for ( i = 0; i < _face.length; i++ )
		{
			var v:Vector3D = m.transform(_world[_face[i]._index[0]]);
			var v1:Vector3D = m.transform(_world[_face[i]._index[1]]);
			var v2:Vector3D = m.transform(_world[_face[i]._index[2]]);
			faces.push((v._z+v1._z+v2._z)/3);
		}

		//ニアクリッピング
		for ( i = 0; i < faces.length; i++ )
		{
			if ( faces[i] < camera._near || faces[i] > camera._far )
			{
				_face[i]._visible = false;
			}
		}

		// Zソート (奥のポリゴンから順番に並べる)
		_index = faces.sort(
			Array.DESCENDING |
			Array.NUMERIC |
			Array.RETURNINDEXEDARRAY);
	}
}
